@@ -20,18 +20,22 @@ import six
 import struct
 import weakref
 
-# We're just re-exporting visualtype_to_c_struct, hence the noqa.
-from .ffi import ffi, C, visualtype_to_c_struct  # noqa
+try:
+    from xcffib._ffi import ffi
+except ImportError:
+    from xcffib.ffi_build import ffi
+
+lib = ffi.dlopen('libxcb.so')
 
 __xcb_proto_version__ = 'placeholder'
 
-X_PROTOCOL = C.X_PROTOCOL
-X_PROTOCOL_REVISION = C.X_PROTOCOL_REVISION
+X_PROTOCOL = lib.X_PROTOCOL
+X_PROTOCOL_REVISION = lib.X_PROTOCOL_REVISION
 
-XCB_NONE = C.XCB_NONE
-XCB_COPY_FROM_PARENT = C.XCB_COPY_FROM_PARENT
-XCB_CURRENT_TIME = C.XCB_CURRENT_TIME
-XCB_NO_SYMBOL = C.XCB_NO_SYMBOL
+XCB_NONE = lib.XCB_NONE
+XCB_COPY_FROM_PARENT = lib.XCB_COPY_FROM_PARENT
+XCB_CURRENT_TIME = lib.XCB_CURRENT_TIME
+XCB_NO_SYMBOL = lib.XCB_NO_SYMBOL
 
 # For xpyb compatibility
 NONE = XCB_NONE
@@ -39,13 +43,13 @@ CopyFromParent = XCB_COPY_FROM_PARENT
 CurrentTime = XCB_CURRENT_TIME
 NoSymbol = XCB_NO_SYMBOL
 
-XCB_CONN_ERROR = C.XCB_CONN_ERROR
-XCB_CONN_CLOSED_EXT_NOTSUPPORTED = C.XCB_CONN_CLOSED_EXT_NOTSUPPORTED
-XCB_CONN_CLOSED_MEM_INSUFFICIENT = C.XCB_CONN_CLOSED_MEM_INSUFFICIENT
-XCB_CONN_CLOSED_REQ_LEN_EXCEED = C.XCB_CONN_CLOSED_REQ_LEN_EXCEED
-XCB_CONN_CLOSED_PARSE_ERR = C.XCB_CONN_CLOSED_PARSE_ERR
-# XCB_CONN_CLOSED_INVALID_SCREEN = C.XCB_CONN_CLOSED_INVALID_SCREEN
-# XCB_CONN_CLOSED_FDPASSING_FAILED = C.XCB_CONN_CLOSED_FDPASSING_FAILED
+XCB_CONN_ERROR = lib.XCB_CONN_ERROR
+XCB_CONN_CLOSED_EXT_NOTSUPPORTED = lib.XCB_CONN_CLOSED_EXT_NOTSUPPORTED
+XCB_CONN_CLOSED_MEM_INSUFFICIENT = lib.XCB_CONN_CLOSED_MEM_INSUFFICIENT
+XCB_CONN_CLOSED_REQ_LEN_EXCEED = lib.XCB_CONN_CLOSED_REQ_LEN_EXCEED
+XCB_CONN_CLOSED_PARSE_ERR = lib.XCB_CONN_CLOSED_PARSE_ERR
+# XCB_CONN_CLOSED_INVALID_SCREEN = lib.XCB_CONN_CLOSED_INVALID_SCREEN
+# XCB_CONN_CLOSED_FDPASSING_FAILED = lib.XCB_CONN_CLOSED_FDPASSING_FAILED
 
 cffi_explicit_lifetimes = weakref.WeakKeyDictionary()
 
@@ -54,6 +58,22 @@ def type_pad(t, i):
     return -i & (3 if t > 4 else t - 1)
 
 
+def visualtype_to_c_struct(vt):
+    # let ffi be a kwarg so cairocffi can pass in its ffi
+    # cfficairo needs an xcb_visualtype_t
+    s = ffi.new("struct xcb_visualtype_t *")
+
+    s.visual_id = vt.visual_id
+    s._class = vt._class
+    s.bits_per_rgb_value = vt.bits_per_rgb_value
+    s.colormap_entries = vt.colormap_entries
+    s.red_mask = vt.red_mask
+    s.green_mask = vt.green_mask
+    s.blue_mask = vt.blue_mask
+
+    return s
+
+
 class Unpacker(object):
 
     def __init__(self, known_max=None):
@@ -123,9 +143,9 @@ class CffiUnpacker(Unpacker):
 
 class MemoryUnpacker(Unpacker):
 
-    def __init__(self, data, fmt):
-        self.buf = struct.pack(fmt, *data)
-        Unpacker.__init__(self, struct.calcsize(fmt))
+    def __init__(self, buf):
+        self.buf = buf
+        Unpacker.__init__(self, len(self.buf))
 
     def _resize(self, increment):
         if self.size + increment > self.known_max:
@@ -133,11 +153,9 @@ class MemoryUnpacker(Unpacker):
         self.size += increment
 
     def copy(self):
-        new = MemoryUnpacker.__new__(MemoryUnpacker)
-        new.buf = self.buf
+        new = MemoryUnpacker(self.buf)
         new.offset = self.offset
         new.size = self.size
-        new.known_max = self.known_max
         return new
 
 
@@ -153,23 +171,23 @@ class XcffibException(Exception):
 
 class ConnectionException(XcffibException):
     REASONS = {
-        C.XCB_CONN_ERROR: (
+        lib.XCB_CONN_ERROR: (
             'xcb connection errors because of socket, '
             'pipe and other stream errors.'),
-        C.XCB_CONN_CLOSED_EXT_NOTSUPPORTED: (
+        lib.XCB_CONN_CLOSED_EXT_NOTSUPPORTED: (
             'xcb connection shutdown because extension not supported'),
-        C.XCB_CONN_CLOSED_MEM_INSUFFICIENT: (
+        lib.XCB_CONN_CLOSED_MEM_INSUFFICIENT: (
             'malloc(), calloc() and realloc() error upon failure, '
             'for eg ENOMEM'),
-        C.XCB_CONN_CLOSED_REQ_LEN_EXCEED: (
+        lib.XCB_CONN_CLOSED_REQ_LEN_EXCEED: (
             'Connection closed, exceeding request length that server '
             'accepts.'),
-        C.XCB_CONN_CLOSED_PARSE_ERR: (
+        lib.XCB_CONN_CLOSED_PARSE_ERR: (
             'Connection closed, error during parsing display string.'),
-        #        C.XCB_CONN_CLOSED_INVALID_SCREEN: (
+        #        lib.XCB_CONN_CLOSED_INVALID_SCREEN: (
         #            'Connection closed because the server does not have a screen '
         #            'matching the display.'),
-        #        C.XCB_CONN_CLOSED_FDPASSING_FAILED: (
+        #        lib.XCB_CONN_CLOSED_FDPASSING_FAILED: (
         #            'Connection closed because some FD passing operation failed'),
     }
 
@@ -251,9 +269,11 @@ class Protobj(object):
     """ Note: Unlike xcb.Protobj, this does NOT implement the sequence
     protocol. I found this behavior confusing: Protobj would implement the
     sequence protocol on self.buf, and then List would go and implement it on
-    List. Additionally, as near as I can tell internally we only need the size
-    of the buffer for cases when the size of things is unspecified. Thus,
-    that's all we save.
+    List.
+
+    Instead, when we need to create a new event from an existing event, we
+    repack that event into a MemoryUnpacker and use that instead (see
+    eventToUnpacker in the generator for more info.)
     """
 
     def __init__(self, unpacker):
@@ -283,7 +303,7 @@ class Union(Protobj):
     @classmethod
     def synthetic(cls, data=[], fmt=""):
         self = cls.__new__(cls)
-        self.__init__(MemoryUnpacker(data, fmt))
+        self.__init__(MemoryUnpacker(struct.pack(fmt, *data)))
         return self
 
 
@@ -350,7 +370,7 @@ class Extension(object):
         xcb_parts[3].iov_base = ffi.NULL
         xcb_parts[3].iov_len = -len(data) & 3  # is this really necessary?
 
-        flags = C.XCB_REQUEST_CHECKED if is_checked else 0
+        flags = lib.XCB_REQUEST_CHECKED if is_checked else 0
 
         seq = self.conn.send_request(flags, xcb_parts + 2, xcb_req)
 
@@ -471,13 +491,19 @@ class Connection(object):
         i = ffi.new("int *")
 
         if fd > 0:
-            self._conn = C.xcb_connect_to_fd(fd, c_auth)
+            self._conn = lib.xcb_connect_to_fd(fd, c_auth)
         elif c_auth != ffi.NULL:
-            self._conn = C.xcb_connect_to_display_with_auth_info(display, c_auth, i)
+            self._conn = lib.xcb_connect_to_display_with_auth_info(display, c_auth, i)
         else:
-            self._conn = C.xcb_connect(display, i)
+            self._conn = lib.xcb_connect(display, i)
         self.pref_screen = i[0]
         self.invalid()
+        self._init_x()
+
+    def _init_x(self):
+        if core is None:
+            raise XcffibException("No core protocol object has been set.  "
+                                  "Did you import xcffib.xproto?")
 
         self.core = core(self)
         self.setup = self.get_setup()
@@ -491,7 +517,7 @@ class Connection(object):
             # We're explicitly not putting this as an argument to the next call
             # as a hack for lifetime management.
             c_ext = key.to_cffi()
-            reply = C.xcb_get_extension_data(self._conn, c_ext)
+            reply = lib.xcb_get_extension_data(self._conn, c_ext)
             self._event_offsets.add(reply.first_event, events)
             self._error_offsets.add(reply.first_error, errors)
 
@@ -501,7 +527,7 @@ class Connection(object):
     def invalid(self):
         if self._conn is None:
             raise XcffibException("Invalid connection.")
-        err = C.xcb_connection_has_error(self._conn)
+        err = lib.xcb_connection_has_error(self._conn)
         if err > 0:
             raise ConnectionException(err)
 
@@ -522,57 +548,70 @@ class Connection(object):
 
     @ensure_connected
     def get_setup(self):
-        s = C.xcb_get_setup(self._conn)
+        self._setup = lib.xcb_get_setup(self._conn)
 
         # No idea where this 8 comes from either, similar complate to the
         # sizeof(xcb_generic_reply_t) below.
-        buf = CffiUnpacker(s, known_max=8 + s.length * 4)
+        buf = CffiUnpacker(self._setup, known_max=8 + self._setup.length * 4)
 
         return setup(buf)
 
+    @ensure_connected
+    def get_screen_pointers(self):
+        """
+        Returns the xcb_screen_t for every screen
+        useful for other bindings
+        """
+        root_iter = lib.xcb_setup_roots_iterator(self._setup)
+
+        screens = [root_iter.data]
+        for i in range(self._setup.roots_len - 1):
+            lib.xcb_screen_next(ffi.addressof((root_iter)))
+            screens.append(root_iter.data)
+        return screens
+
     @ensure_connected
     def wait_for_event(self):
-        e = C.xcb_wait_for_event(self._conn)
-        e = ffi.gc(e, C.free)
+        e = lib.xcb_wait_for_event(self._conn)
+        e = ffi.gc(e, lib.free)
         self.invalid()
         return self.hoist_event(e)
 
     @ensure_connected
     def poll_for_event(self):
-        e = C.xcb_poll_for_event(self._conn)
+        e = lib.xcb_poll_for_event(self._conn)
         self.invalid()
         if e != ffi.NULL:
             return self.hoist_event(e)
         else:
             return None
 
-    @ensure_connected
     def has_error(self):
-        return C.xcb_connection_has_error(self._conn)
+        return lib.xcb_connection_has_error(self._conn)
 
     @ensure_connected
     def get_file_descriptor(self):
-        return C.xcb_get_file_descriptor(self._conn)
+        return lib.xcb_get_file_descriptor(self._conn)
 
     @ensure_connected
     def get_maximum_request_length(self):
-        return C.xcb_get_maximum_request_length(self._conn)
+        return lib.xcb_get_maximum_request_length(self._conn)
 
     @ensure_connected
     def prefetch_maximum_request_length(self):
-        return C.xcb_prefetch_maximum_request_length(self._conn)
+        return lib.xcb_prefetch_maximum_request_length(self._conn)
 
     @ensure_connected
     def flush(self):
-        return C.xcb_flush(self._conn)
+        return lib.xcb_flush(self._conn)
 
     @ensure_connected
     def generate_id(self):
-        return C.xcb_generate_id(self._conn)
+        return lib.xcb_generate_id(self._conn)
 
     def disconnect(self):
         self.invalid()
-        return C.xcb_disconnect(self._conn)
+        return lib.xcb_disconnect(self._conn)
 
     def _process_error(self, c_error):
         self.invalid()
@@ -584,14 +623,14 @@ class Connection(object):
     @ensure_connected
     def wait_for_reply(self, sequence):
         error_p = ffi.new("xcb_generic_error_t **")
-        data = C.xcb_wait_for_reply(self._conn, sequence, error_p)
-        data = ffi.gc(data, C.free)
+        data = lib.xcb_wait_for_reply(self._conn, sequence, error_p)
+        data = ffi.gc(data, lib.free)
 
         try:
             self._process_error(error_p[0])
         finally:
             if error_p[0] != ffi.NULL:
-                C.free(error_p[0])
+                lib.free(error_p[0])
 
         if data == ffi.NULL:
             # No data and no error => bad sequence number
@@ -607,7 +646,7 @@ class Connection(object):
         cookie = ffi.new("xcb_void_cookie_t [1]")
         cookie[0].sequence = sequence
 
-        err = C.xcb_request_check(self._conn, cookie[0])
+        err = lib.xcb_request_check(self._conn, cookie[0])
         self._process_error(err)
 
     def hoist_event(self, e):
@@ -626,7 +665,7 @@ class Connection(object):
 
     @ensure_connected
     def send_request(self, flags, xcb_parts, xcb_req):
-        return C.xcb_send_request(self._conn, flags, xcb_parts, xcb_req)
+        return lib.xcb_send_request(self._conn, flags, xcb_parts, xcb_req)
 
 
 # More backwards compatibility
@@ -641,9 +680,17 @@ class Response(Protobj):
         # These (and the ones in Reply) aren't used internally and I suspect
         # they're not used by anyone else, but they're here for xpyb
         # compatibility.
-        resp = unpacker.cast("xcb_generic_event_t *")
-        self.response_type = resp.response_type
-        self.sequence = resp.sequence
+        #
+        # In some cases (e.g. creating synthetic events from memory), we don't
+        # have the sequence number (since the event was fake), so only try to
+        # get these attributes if we are really using a cffi buffer.
+        if isinstance(unpacker, CffiUnpacker):
+            resp = unpacker.cast("xcb_generic_event_t *")
+            self.response_type = resp.response_type
+            self.sequence = resp.sequence
+        else:
+            self.response_type = None
+            self.sequence = None
 
 
 class Reply(Response):
@@ -709,3 +756,17 @@ def pack_list(from_, pack_type):
             else:
                 buf.write(item)
         return buf.getvalue()
+
+
+def wrap(ptr):
+    c_conn = ffi.cast('xcb_connection_t *', ptr)
+    conn = Connection.__new__(Connection)
+    conn._conn = c_conn
+    conn._init_x()
+    conn.invalid()
+
+    # ptr owns the memory for c_conn, even after the cast
+    # we should keep it alive
+    cffi_explicit_lifetimes[conn] = ptr
+
+    return conn
